from pathlib import Path
from typing import List, Tuple

from bs4 import BeautifulSoup
import openai


def extract_predefined_entities_text(input_html: str, remove_entity: bool = False):
    input_html = BeautifulSoup(input_html, features="html.parser")
    # extract predefined entities
    predefined_enities = []
    for entity in input_html.find_all("mark"):
        predefined_enities.append(entity.get_text())
        if remove_entity:
            entity.decompose()
    # extract text
    text = ""
    for paragraph in input_html.find_all("p"):
        text += paragraph.get_text() + "\n"
    return predefined_enities, text.strip()


def get_html_entity_dummy(input_html: str, prompt: str = "") -> Tuple[List[str], List[List[str]]]:
    predefined_entities, text = extract_predefined_entities_text(input_html, remove_entity=True)
    words = text.split(" ")
    entities = predefined_entities
    relations = []
    for word in words:
        entity = word.strip()
        if entity == "":
            continue
        entities.append(entity)
    for idx, entity in enumerate(entities):
        if idx == 0:
            continue
        relations.append([entities[idx-1], entity])
    return entities, relations


def get_html_entity_gpt3(input_html: str, prompt: str = "") -> Tuple[List[str], List[List[str]]]:
    input = BeautifulSoup(input_html, features="html.parser").get_text()
    token_file = Path(Path.home(), ".nbx/openai_token")
    # @TODO Add warning if token not configured
    with open(token_file, 'r') as f:
        openai.api_key = f.read()

    if prompt == "":
        prompt = f"""
Extract the entities of the input in the appearance order and separate by comma.

Input: "The stock market is in an recession now"
Entities: stock market, recession

Input: "{input}"
Entities:
"""
    prompt = prompt.strip()

    response = openai.Completion.create(
        model="text-davinci-002",
        prompt=prompt,
        temperature=0,
        max_tokens=60,
        top_p=1,
        frequency_penalty=0.5,
        presence_penalty=0
    )
    texts = response["choices"][0]["text"].split(",")
    entities = []
    for text in texts:
        if text.strip() == "":
            continue
        entities.append(text.strip())
    relations = []
    for idx, entity in enumerate(entities):
        if idx == 0:
            continue
        relations.append([entities[idx-1], entity])
    return entities, relations



import os
from pathlib import Path
from typing import List, Tuple, Callable

import pinecone
from sentence_transformers import SentenceTransformer

from nbx.note import Note, Record

MODEL_CKPT = "flax-sentence-embeddings/all_datasets_v3_mpnet-base"
SEARCH_INDEX_NAME = "search"
TIMEOUT_SECONDS = 300


def find_match(
    query: str,
    pinecone_token: str,
    notes_dir: Path,
    fallback_func: Callable,
    top_k: int = 10
) -> Tuple[List[Record], List[Note]]:
    if pinecone_token == "":
        print("[Error] Pinecone token not set, please set with: nbx config pinecone")
        return fallback_func()

    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    model = SentenceTransformer(MODEL_CKPT)
    xq = model.encode(query).tolist()

    pinecone.init(api_key=pinecone_token, environment='us-west1-gcp')
    try:
        index = pinecone.Index(SEARCH_INDEX_NAME)
        results = index.query(xq, top_k=top_k, includeMetadata=True)
    except Exception:
        print("[Warning] Indexing in progress, fall back to regex search")
        return fallback_func()

    record_ids = []
    for match in results["matches"]:
        record_ids.append(match['metadata']["record_id"])

    res_records = []
    res_notes = []
    for record_id in record_ids:
        timestamp = record_id.split("#")[0]
        record_idx = int(record_id.split("#")[1])
        note_file_path = Path(notes_dir, f"{timestamp}.json")
        note = Note(note_file_path)
        record = note.get_sections()[record_idx]
        res_records.append(record)
        res_notes.append(note)

    return res_records, res_notes


# https://www.pinecone.io/docs/examples/basic-hybrid-search/
def index_notes(notes: List[Note], pinecone_token: str) -> None:
    # get embeddings of each record
    encoder = SentenceTransformer(MODEL_CKPT)
    contents = []
    record_ids = []
    for note in notes:
        records = note.get_sections()
        for record in records:
            contents.append(record.content)
            record_ids.append(record.id)
    embeddings = encoder.encode(contents)
    # upsert to pinecone
    print("Connecting to Pinecone")
    pinecone.init(api_key=pinecone_token, environment='us-west1-gcp')
    if SEARCH_INDEX_NAME in pinecone.list_indexes():
        print("Deleting old search index")
        pinecone.delete_index(SEARCH_INDEX_NAME, timeout=TIMEOUT_SECONDS)
    print("Creating new search index")
    pinecone.create_index(
        name=SEARCH_INDEX_NAME,
        dimension=embeddings.shape[1],
        timeout=TIMEOUT_SECONDS
    )
    index = pinecone.Index(SEARCH_INDEX_NAME)
    upserts = []
    for idx, (embedding, record_id) in enumerate(zip(embeddings, record_ids)):
        upserts.append((str(idx), embedding.tolist(), {"record_id": record_id}))
    print("Upserting embeddings")
    status = index.upsert(vectors=upserts)
    upserted_count = status["upserted_count"]
    print(f"Indexed records: {upserted_count}")



import os
from pathlib import Path

import typer
import html2text
from rich.panel import Panel
from rich.prompt import Prompt, Confirm
from rich.console import Console
from rich.markdown import Markdown

from nbx import nbx, util, config, tui


TUI_MENU_COMMANDS = ["nbx", "nbx add", "nbx edit <index>",
                     "nbx search <query>", "nbx insight"]

cli = typer.Typer()
console = Console()


#######################################################################################
# Insight CLI

insight_cli = typer.Typer()


@insight_cli.command("add")
def add_insight():
    """
    Learn something from the note base
    """
    task = Prompt.ask("Choose a task type", choices=["Analyze a Note"], default="Analyze a Note")
    console.print(f"Not Yet Implemented: {task}")


cli.add_typer(insight_cli, name="insight")

#######################################################################################
# Config CLI

config_cli = typer.Typer()


@config_cli.callback(invoke_without_command=True)
def config_main(ctx: typer.Context):
    """
    Configure NBX
    """
    # only run without a command specified
    if ctx.invoked_subcommand is not None:
        return

    nbx_config = config.load_config()
    notes_dir = nbx_config["notes_dir"]
    serach_method = nbx_config["search_method"]
    config_subcommands = [
        "nbx config search", "nbx config storage",
        "nbx config openai", "nbx config pinecone"
    ]
    subtitle = tui.get_panel_subtitle(config_subcommands)
    content = f"""
[bold cyan]Notes Repo:[/bold cyan] [yellow]{str(notes_dir)}[/yellow]
[bold cyan]Search Method:[/bold cyan] [yellow]{serach_method}[/yellow]
    """
    dashboard = Panel(
        content.strip(), title="NBX Configuration", subtitle=subtitle
    )
    console.print(dashboard)


@config_cli.command("storage")
def config_storage():
    """
    Configure NBX's storage
    """
    notes_dir = Prompt.ask("Enter repo path", default=os.fspath(config.DEFAULT_NOTES_DIR)).strip()
    notes_dir = Path(notes_dir)
    notes_dir.mkdir(parents=True, exist_ok=True)
    config.set_notes_dir(notes_dir)


@config_cli.command("search")
def config_search_method():
    """
    Configure search method
    """
    from nbx.search import SEARCH_OPTIONS
    nbx_config = config.load_config()
    method = Prompt.ask(
        "Enter your preferred search method",
        choices=SEARCH_OPTIONS, default=nbx_config["search_method"]
    ).strip()
    config.set_search_method(method)
    nbx.index_all_notes(background=True)


@config_cli.command("openai")
def config_openai():
    """
    Configure OpenAI API token
    """
    token = Prompt.ask("Enter your OpenAI API token").strip()
    verify_resp = util.verify_openai_token(token)
    if verify_resp != "OK":
        console.print(verify_resp)
        return
    config.set_openai_token(token)


@config_cli.command("pinecone")
def config_pincone():
    """
    Configure Pincone API token
    """
    token = Prompt.ask("Enter your Pinecone API token").strip()
    verify_resp = util.verify_pinecone_token(token)
    if verify_resp != "OK":
        console.print(verify_resp)
        return
    config.set_pinecone_token(token)


cli.add_typer(config_cli, name="config")

#######################################################################################
# Main CLI


@cli.command("add")
def add(message: str = typer.Option("", "--message", "-m", help="Message to add")):
    """
    Add a new note
    """
    if message == "":
        from nbx import app
        app.add_note()
        return

    nbx.create_note(f"<p>{message}</p>")
    nbx.index_all_notes(background=True)


@cli.command("edit")
def edit(index: int):
    """
    Edit the note with <index>
    """
    notes = nbx.get_all_notes()
    if len(notes) <= index:
        console.print("[Warning]: Note doesn't exist")
        return

    from nbx import app
    app.edit_note(notes[index])


@cli.command("search")
def search(
    query: str,
    edit_index: int = typer.Option(
        -1, "--edit", "-e", help="edit the search result with the index"
    ),
    peek_index: int = typer.Option(
        -1, "--peek", "-p", help="peek the serach result with the index"
    )
):
    """
    Search notes with query
    """
    all_notes = nbx.get_all_notes()
    if len(all_notes) == 0:
        console.print("[Warning]: No note exist")
        return

    records, notes = nbx.search_notes(all_notes, query)

    if len(records) > 0:
        if edit_index >= 0 and edit_index < len(notes):
            from nbx import app
            app.edit_note(notes[edit_index])
            return
        if peek_index >= 0 and peek_index < len(notes):
            from nbx import app
            app.edit_note(notes[peek_index], read_only=True)
            return
        panel_content = tui.get_panel_records(records, notes)
    else:
        panel_content = "No matching"

    dashboard = Panel(panel_content, subtitle=tui.get_panel_subtitle(TUI_MENU_COMMANDS))
    console.print(dashboard)


@cli.command("archive")
def archive(index: int):
    """
    Archive the note with <index>
    """
    notes = nbx.get_all_notes()
    if len(notes) <= index:
        console.print("[Warning]: Note doesn't exist")
        return
    note = notes[index]
    prompt = note.get_title()
    if not prompt:
        prompt = note.get_headline()
    if len(prompt) > 120:
        prompt = prompt[:117]
        prompt += "..."
    confirmed = Confirm.ask(prompt)
    if confirmed:
        note.archive()
        nbx.index_all_notes(background=True)


@cli.command("index")
def index():
    """
    Index (retrain) the note base
    """
    all_notes = nbx.get_all_notes()
    console.print("Indexing the note base...")
    nbx.index_notes(all_notes)
    console.print("Done")


@cli.command("log")
def log():
    """
    Show the log of changes in nbx
    """
    logs = ["log1", "log2"]
    log_content = tui.get_log_content(logs)
    panel_subtitle = tui.get_panel_subtitle(TUI_MENU_COMMANDS)
    log_panel = Panel(log_content, title="NBX Logs",
                      subtitle=panel_subtitle)
    console.print(log_panel)


@cli.command("peek")
def peek(
    index: int = typer.Argument(..., help="index of the note"),
    gui: bool = typer.Option(
        False, "--gui", "-g", help="open the note with GUI in browser"
    ),
):
    """
    Peek a note by <index>
    """
    notes = nbx.get_all_notes()
    if len(notes) <= index:
        console.print("[Warning]: Note doesn't exist")
    else:
        if gui:
            from nbx import app
            app.edit_note(notes[index], read_only=True)
        else:
            note = notes[index]
            markdown_text = html2text.html2text(note.html, bodywidth=0)
            md = Markdown(markdown_text)
            console.print(md)


@cli.command("status")
def status():
    """
    Status of the nbx note base
    """
    console.print("NBX status")


@cli.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    gui: bool = typer.Option(
        False, "--gui", "-g", help="open the note with GUI in browser"
    )
):
    """
    Default behavior
    """
    # only run without a command specified
    if ctx.invoked_subcommand is not None:
        return

    notes = nbx.get_all_notes()

    if gui:
        pass
    else:
        # fake saved queries
        saved_queries = ["NBX-TODO", "Micro-SAAS", "GPT3"]
        panel_title = tui.get_panel_title(saved_queries)
        panel_subtitle = tui.get_panel_subtitle(TUI_MENU_COMMANDS)
        panel_content = tui.get_panel_notes(notes)
        panel = Panel(panel_content, title=panel_title, subtitle=panel_subtitle)
        console.print(panel)

    # if gui:
    #     from nbx import app
    #     # extract only title string
    #     note_titles = [note_title[0] for note_title in note_titles]
    #     note_infos = [
    #         {"filename": file, "title": title} for file, title in zip(note_files, note_titles)
    #     ]
    #     app.list_notes(note_infos)
    # else:
    #     # fake saved queries
    #     saved_queries = ["NBX-TODO", "Micro-SAAS", "GPT3"]

    #     # format data into rich dashboard desired format
    #     panel_title = tui.get_panel_title(saved_queries)
    #     panel_subtitle = tui.get_panel_subtitle(TUI_MENU_COMMANDS)
    #     panel_content = tui.get_panel_content(note_titles, note_tags)
    #     dashboard = Panel(panel_content, title=panel_title, subtitle=panel_subtitle)
    #     console.print(dashboard)



import shutil
import tempfile
import time
from pathlib import Path

from nbx.note import Note, create_note_file

NOTES_HTML = [
    "./notes/title_heading_content.html",
    "./notes/only_heading.html",
    "./notes/heading_heading.html"
]

KEYWORD_QUERY = "title_heading_content"
NON_EXIST_QUERY = "shouldn't exist"


def test_search_regex():
    from nbx.search.regex import find_match

    notes_dir = Path(tempfile.mkdtemp())

    # create notes_file
    notes = []
    for note_html in NOTES_HTML:
        with open(note_html) as file:
            note_path = create_note_file(file.read(), notes_dir)
        note = Note(note_path)
        notes.append(note)
        time.sleep(1.5)  # sleep 1.5 second, otherwise notes will have the same timestamp

    # test find_match
    res_sections, res_notes = find_match(notes, KEYWORD_QUERY)
    assert len(res_sections) > 0
    # QUERY is an excerpt from the first record of the title_heading_content
    first_note = res_notes[0]
    assert res_sections[0].id == f"{first_note.last_update}#0"

    # this query shouldn't return any matched result
    res_sections, res_notes = find_match(notes, NON_EXIST_QUERY)
    assert len(res_sections) == 0

    # cleanup the temporary notes directory
    shutil.rmtree(notes_dir)


if __name__ == "__main__":
    test_search_regex()



from typing import List, Tuple

from txtai.embeddings import Embeddings

from nbx.note import Note, Record


MODEL_CKPT = "sentence-transformers/nli-mpnet-base-v2"


# https://github.com/neuml/txtai/blob/master/examples/similarity.py
def find_match(notes: List[Note], query: str) -> Tuple[List[Record], List[Note]]:
    embedding = Embeddings({"path": MODEL_CKPT})
    res_records = []
    res_notes = []
    data = []
    record_note = []
    for note in notes:
        for record in note.get_sections():
            text_data = record.heading + record.content
            data.append(text_data)
            record_note.append([record, note])
    results = embedding.similarity(query, data)
    results = results[:10] if len(results) > 10 else results
    for result in results:
        res_id = result[0]
        res_records.append(record_note[res_id][0])
        res_notes.append(record_note[res_id][1])
    return res_records, res_notes



import shutil
import tempfile
import time
from pathlib import Path

from nbx.note import Note, create_note_file

NOTE_WITH_TITLE = """<h1>Test note with title</h1><p>This is a description right after the note title and before the first h2 heading. This section performs as an abstraction or introduction area of the entire note.</p><h2>First section</h2><p>This is the first section that user can jot down more detailed information that is contained as a section. Each section should be self-containable in terms of the logic. Then the user can restitch the section with other sections into a full complete note without much modification to the content.</p><h2>Second section</h2><p>Here is the second section of the note. As mentioned earlier, each section should be self-containable in terms of logic. This second section should talk about some thing different from the first section.</p><p>Each section may contain more than one paragraph. As long as the entire section is complete in terms of the logic. </p><h2>Section with bullet points</h2><p>A second could also contain contents other than test, for example bullet points</p><ol><li><p>The first item in the list</p></li><li><p>The second item in the list</p></li></ol>"""  # noqa: E501
NOTE_WITHOUT_TITLE = """<p>Note without title. Sometime there are unformalized notes that don't really have a title. We should be able to put this kind of notes also part of the searchable content.</p><h2>First section</h2><p>Even though the note doesn't have a title, it might still have sections.</p>"""  # noqa: E501
QUERY = "We should be able to put this kind of notes also part of the searchable content"


def test_search_faiss():
    from nbx.search.faiss import index_notes, find_match

    # test index_notes
    notes_dir = Path(tempfile.mkdtemp())
    note_with_title_path = create_note_file(NOTE_WITH_TITLE, notes_dir)
    note_with_title = Note(note_with_title_path)
    time.sleep(1.5)  # sleep 1.5 second, otherwise the generated notes will have the same timestamp
    note_without_title_path = create_note_file(NOTE_WITHOUT_TITLE, notes_dir)
    note_without_title = Note(note_without_title_path)
    notes = [note_with_title, note_without_title]
    index_dir = Path(notes_dir, "faiss")
    index_notes(notes, index_dir)

    records_cnt = 0
    for note in notes:
        records_cnt += len(note.get_sections())

    # test find_match
    res_records, res_notes = find_match(QUERY, index_dir, notes_dir, top_k=records_cnt)

    assert len(res_records) == records_cnt
    assert len(res_notes) == len(res_records)
    # QUERY is an excerpt from the first record of the note_without_title
    assert res_records[0].id == f"{note_without_title.last_update}#0"
    # cleanup the temporary notes directory
    shutil.rmtree(notes_dir)


if __name__ == "__main__":
    test_search_faiss()



from pathlib import Path
from typing import Any, List, Tuple
import uuid

from nbx import config, util
from nbx.note import Note, Record, create_note_file

ALL_NOTES = []


def get_all_notes() -> List[Note]:
    nbx_config = config.load_config()
    notes_dir = Path(nbx_config["notes_dir"])
    note_files = notes_dir.glob("*.json")
    note_files = sorted(note_files, reverse=True)
    notes = []
    for note_file in note_files:
        notes.append(Note(note_file))
    return notes


def create_note(content: str) -> None:
    nbx_config = config.load_config()
    notes_dir = Path(nbx_config["notes_dir"])
    create_note_file(content, notes_dir)


def search_notes(notes: List[Note], query: str) -> Tuple[List[Record], List[Note]]:
    from nbx import search
    nbx_config = config.load_config()
    search_method = nbx_config["search_method"]
    return search.search_notes(notes, query, search_method)


def find_relevant_records(content: str) -> List[List[str]]:
    """
    Find abstracts of relevant note based on content around the cursor.

    Output: [{"heading": "Note Title", "headingId": "123456", "content": "Note Paragraph"},...]
    """
    print(f"[Current User Input] {content}")
    global ALL_NOTES
    if not ALL_NOTES:
        ALL_NOTES = get_all_notes()
    if len(content) < 10:
        return []
    try:
        records, notes = search_notes(ALL_NOTES, content)
    except Exception:
        records = []
    result = []
    for record in records:
        result.append(vars(record))
    return result


def add_image(image: Any) -> str:
    nbx_config = config.load_config()
    notes_dir = Path(nbx_config["notes_dir"])
    images_dir = Path(notes_dir, "images")
    images_dir.mkdir(parents=True, exist_ok=True)
    image_name = f"{str(uuid.uuid4())}.png"
    image_path = Path(notes_dir, "images", image_name)
    util.save_base64(image, image_path)
    return image_name


def get_image_path(filename: str) -> Path:
    nbx_config = config.load_config()
    notes_dir = Path(nbx_config["notes_dir"])
    image_path = Path(notes_dir, "images", filename)
    return image_path


def index_notes(notes: List[Note]) -> None:
    from nbx import search
    nbx_config = config.load_config()
    search_method = nbx_config["search_method"]
    search.index_notes(notes, search_method)


def index_all_notes(background: bool = False) -> Note:
    if background:
        import subprocess
        subprocess.Popen(["nbx", "index"], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
    else:
        all_notes = get_all_notes()
        index_notes(all_notes)



import os
from datetime import datetime
from typing import List, Tuple
from pathlib import Path

from whoosh.fields import Schema, TEXT, ID
from whoosh.index import create_in, open_dir
from whoosh.qparser import QueryParser

from nbx.note import Note, Record


def find_match(
    query: str,
    index_dir: Path,
    notes_dir: Path,
    top_k: int = 10
) -> Tuple[List[Record], List[Note]]:
    if not index_dir.exists():
        print("No index found, please index the notebase first: nbx index")
        return [], []

    # get the index with the latest timestamp
    indices = [f.path for f in os.scandir(index_dir) if f.is_dir() and f.name.startswith("index_")]
    indices = sorted(indices, reverse=True)
    if len(indices) == 0:
        print("No index found, please index the notebase first: nbx index")
        return [], []

    ix = open_dir(indices[0])
    with ix.searcher() as searcher:
        parser = QueryParser("content", ix.schema)
        myquery = parser.parse(query)
        results = searcher.search(myquery, limit=top_k)
        # convert results to records & notes
        res_records = []
        res_notes = []
        loaded_notes = {}
        loaded_records = {}
        for result in results:
            id = result["id"]
            timestamp = id.split("#")[0]
            record_idx = int(id.split("#")[1])
            note_file_path = Path(notes_dir, f"{timestamp}.json")
            if note_file_path not in loaded_notes:
                note = Note(note_file_path)
                loaded_notes[note_file_path] = note
                loaded_records[note_file_path] = note.get_sections()
            note = loaded_notes[note_file_path]
            record = loaded_records[note_file_path][record_idx]
            res_records.append(record)
            res_notes.append(note)

    return res_records, res_notes


# https://whoosh.readthedocs.io/en/latest/quickstart.html
def index_notes(notes: List[Note], index_dir: Path) -> None:
    index_dir.mkdir(parents=True, exist_ok=True)
    timestamp = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
    current_index_dir = Path(index_dir, f"index_{timestamp}")
    current_index_dir.mkdir(parents=True, exist_ok=True)

    # create the index
    schema = Schema(id=ID(stored=True), heading=TEXT, content=TEXT)
    ix = create_in(current_index_dir, schema)

    # add documents
    writer = ix.writer()
    for note in notes:
        for record in note.get_sections():
            writer.add_document(
                id=record.id,
                heading=record.heading.lower(),
                content=record.content.lower()
            )
    writer.commit()



import os
from typing import List
from nbx.note import Note, Record

from rich.text import Text
from rich.console import group

from nbx import util

SEARCH_RESULT_TITLE_TEXT_MAX_LENGTH_PERCENTAGE = 0.9
TITLE_TEXT_MAX_LENGTH_PERCENTAGE = 0.8
TAGS_STRING_TEXT_MAX_LENGTH_PERCENTAGE = 0.15


@group()
def get_log_content(logs: List[str]) -> Text:
    for log in logs:
        yield log


@group()
def get_panel_records(records: List[Record], notes: List[Note]) -> Text:
    terminal_width = os.get_terminal_size().columns
    for idx, (record, note) in reversed(list(enumerate(zip(records, notes)))):
        title = note.get_title().strip()
        heading = record.heading.strip()
        content = record.content.strip()
        # truncage content if it takes more than 3 lines
        lines = content.split("\n")
        if len(lines) > 3:
            content = "\n".join(lines[:3])
        if len(content) > terminal_width * 2:
            content = content[:terminal_width * 2]
        # note title or headline
        styled_title = f"[[blue]{idx}[/blue]]" + " " * \
            (len(str(len(notes))) - len(str(idx)) + 1)
        if title:
            styled_title += f"[dim cyan]{title.strip()}[/dim cyan] " + "[blue]❯ [/blue]"
        if heading:
            styled_title += f"{heading}"
        else:
            styled_title += f"{content}"
        record_text = Text.from_markup(styled_title)
        record_text.truncate(
            int(terminal_width * SEARCH_RESULT_TITLE_TEXT_MAX_LENGTH_PERCENTAGE),
            overflow="ellipsis"
        )
        if content:
            record_text.append_text(Text.from_markup(f"\n[dim]{content}[/dim]"))
        yield record_text


@group()
def get_panel_notes(notes: List[Note]) -> Text:
    PADDING_MAGIC_NUM = 4
    terminal_width = os.get_terminal_size().columns
    for idx, note in reversed(list(enumerate(notes))):
        title = note.get_title()
        headline = note.get_headline()
        # note title or headline
        styled_title = f"[[blue]{idx}[/blue]]" + " " * \
            (len(str(len(notes))) - len(str(idx)) + 1)
        if not title == "":
            styled_title += title
        else:
            styled_title += f"[dim white]{headline}[/dim white]"
        title_text = Text.from_markup(styled_title)
        # note tags
        tags = note.get_tags()
        tags_str = ""
        for tag in tags:
            tags_str += fr" [{tag}]"
        tags_text = Text.from_markup(f"[dim white]{tags_str}[/dim white]")
        # format note title & tags texts
        title_text.truncate(
            int(terminal_width * TITLE_TEXT_MAX_LENGTH_PERCENTAGE), overflow="ellipsis")
        tags_text.truncate(
            int(terminal_width * TAGS_STRING_TEXT_MAX_LENGTH_PERCENTAGE), overflow="ellipsis")
        magic_num = PADDING_MAGIC_NUM
        if util.lang_detect(title_text.plain) == "zh":
            zh_char_cnt = util.count_zh_char(title_text.plain)
            magic_num = PADDING_MAGIC_NUM + zh_char_cnt
        white_space_cnt = terminal_width - title_text._length - \
            tags_text._length - magic_num
        content_text = title_text
        content_text.pad_right(white_space_cnt)
        content_text.append_text(tags_text)
        yield content_text


def get_panel_title(tags: List[str] = [], selected_tag_idxs: List[int] = []) -> str:
    title_str = "Saved Query: " if len(tags) > 0 else ""
    markup_start = "[u blue]"
    markup_end = "[/u blue]"
    separator = "[dim white]·[/dim white]"
    for idx, tag in enumerate(tags):
        # captialize each word in tag to avoid confusion with styling text
        tag = f"[{tag if tag[0].isupper() else tag.title()}]"
        if idx > 0:
            title_str += f" {separator}"
        if idx in selected_tag_idxs:
            tag = f"{markup_start}{tag}{markup_end}"
        title_str += f" {tag}"
    return title_str.strip()


def get_panel_subtitle(commands: List[str]) -> str:
    subtitle_str = "[dim white]❯[/dim white]"
    separator = "[dim blue]·[/dim blue]"
    for idx, command in enumerate(commands):
        if idx > 0:
            subtitle_str += f" {separator}"
        subtitle_str += f" [dim white]{command}[/dim white]"
    return subtitle_str.strip()



import os
from pathlib import Path

from nbx.util import load_json, set_json_value, load_json_value

DEFAULT_NOTES_DIR = Path(Path.home(), ".nbx/notes")


def load_config():
    # default config
    config = {
        "notes_dir": DEFAULT_NOTES_DIR,
        "search_method": "regex"
    }
    config_file = Path(Path.home(), ".nbx/config.json")
    if config_file.exists():
        config.update(load_json(config_file))
    return config


def set_notes_dir(notes_dir: Path):
    # write to the configuration
    config_dir = Path(Path.home(), ".nbx")
    if not config_dir.exists():
        config_dir.mkdir(parents=True)
    config_file = Path(config_dir, "config.json")
    config_file.touch(exist_ok=True)
    set_json_value(config_file, "notes_dir", os.fspath(notes_dir))


def set_search_method(method: str):
    config_dir = Path(Path.home(), ".nbx")
    if not config_dir.exists():
        config_dir.mkdir(parents=True)
    config_file = Path(config_dir, "config.json")
    set_json_value(config_file, "search_method", method)


def set_openai_token(token: str):
    config_dir = Path(Path.home(), ".nbx")
    if not config_dir.exists():
        config_dir.mkdir(parents=True)
    config_file = Path(config_dir, "config.json")
    set_json_value(config_file, "openai_token", token)


def set_pinecone_token(token: str):
    config_dir = Path(Path.home(), ".nbx")
    if not config_dir.exists():
        config_dir.mkdir(parents=True)
    config_file = Path(config_dir, "config.json")
    set_json_value(config_file, "pinecone_token", token)


def load_pinecone_token() -> str:
    config_dir = Path(Path.home(), ".nbx")
    if not config_dir.exists():
        return ""
    config_file = Path(config_dir, "config.json")
    return load_json_value(config_file, "pinecone_token", "")



import glob
import shutil
import tempfile
import os
import time
from pathlib import Path

from bs4 import BeautifulSoup

from nbx.note import Note, create_note_file

NOTE_HTML_FILES = [
    "./notes/content_heading.html",
    "./notes/heading_content_heading.html",
    "./notes/heading_content.html",
    "./notes/heading_heading.html",
    "./notes/only_content_lists.html",
    "./notes/only_content.html",
    "./notes/only_heading.html",
    "./notes/only_title.html",
    "./notes/title_content_heading_heading.html",
    "./notes/title_content_heading.html",
    "./notes/title_content_lists.html",
    "./notes/title_content.html",
    "./notes/title_heading_content_heading_heading.html",
    "./notes/title_heading_content.html"
]

NOTE_TITLES = [
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "Title: only_title",
    "Title: title_content_heading_heading",
    "Title: title_content_heading",
    "Title: title_content_lists",
    "Title: title_content",
    "Title: title_heading_content_heading_heading",
    "Title: title_heading_content"
]

NOTE_SECTION_CNTS = [2, 2, 1, 2, 1, 1, 1, 1, 3, 2, 1, 1, 4, 2]

NOTE_HEADLINES = [
    "This is the paragraph of the content",
    "A section heading",
    "A section heading",
    "A section heading",
    "list content item 1\nlist content item 2\nlist content item 3",
    "The note could only contain a short paragraph as the content. Normally this\n  kind of notes is added via nbx add -m",  # noqa: E501
    "Only a section heading",
    "Title: only_title",
    "Title: title_content_heading_heading",
    "Title: title_content_heading",
    "Title: title_content_lists",
    "Title: title_content",
    "Title: title_heading_content_heading_heading",
    "Title: title_heading_content"
]


def _note_update_body(note_file: Path):
    # setup
    notes_dir = Path(tempfile.mkdtemp())
    with open(note_file) as file:
        note_path = create_note_file(file.read(), notes_dir)
    note = Note(note_path)
    original_title = note.get_title()
    original_body = BeautifulSoup(note.get_body(), features="html.parser").prettify()
    updated_body = BeautifulSoup(
        "<p>updated paragraph body by test_note_update_body</p>",
        features="html.parser"
    ).prettify()
    note.update_body(updated_body)
    # check whether the temporary html file exists
    tmp_html_file = Path(notes_dir, ".tmp", f"{note.last_update}.html")
    assert tmp_html_file.exists(), "Temporary html file doesn't exist after update"
    # check whether the original disk note file unmodified
    original_note = Note(note_path, cache=False)
    modified_err_msg = f"Original disk file got modifed while updating {note_file}"
    assert original_note.get_title() == original_title, modified_err_msg
    original_note_body = BeautifulSoup(original_note.get_body(), features="html.parser").prettify()
    assert original_note_body == original_body, modified_err_msg
    # check whether newly loaded note has the updated body
    new_note = Note(note_path)
    load_err_msg = f"Newly loaded note doesn't have the update while updating {note_file}"
    assert new_note.get_title() == original_title, load_err_msg
    assert new_note.get_body() == updated_body, load_err_msg
    # check whether the tempoary html file has the right title
    with open(tmp_html_file) as html_file:
        tmp_html = BeautifulSoup(html_file.read(), features="html.parser")
    if tmp_html.find("h1"):
        assert tmp_html.find("h1").text.strip() == original_title
        tmp_html.find("h1").decompose()
    assert tmp_html.prettify() == updated_body


def test_note_update_body():
    for note_file in NOTE_HTML_FILES:
        _note_update_body(note_file)


def _note_update_title_titled(note_file: Path):
    # setup
    notes_dir = Path(tempfile.mkdtemp())
    with open(note_file) as file:
        note_path = create_note_file(file.read(), notes_dir)
    note = Note(note_path)
    original_title = note.get_title()
    original_body = BeautifulSoup(note.get_body(), features="html.parser").prettify()
    updated_title = "updated title by test_note_update_title"
    note.update_title(updated_title)
    # check whether the temporary html file exists
    tmp_html_file = Path(notes_dir, ".tmp", f"{note.last_update}.html")
    assert tmp_html_file.exists(), "Temporary html file doesn't exist after update"
    # check whether the original disk note file unmodified
    original_note = Note(note_path, cache=False)
    modified_err_msg = f"Original disk file got modifed while updating {note_file}"
    assert original_note.get_title() == original_title, modified_err_msg
    original_note_body = BeautifulSoup(original_note.get_body(), features="html.parser").prettify()
    assert original_note_body == original_body, modified_err_msg
    # check whether newly loaded note has the updated title
    new_note = Note(note_path)
    load_err_msg = f"Newly loaded note doesn't have the update while updating {note_file}"
    assert new_note.get_title() == updated_title, load_err_msg
    assert new_note.get_body() == original_body, load_err_msg
    # check whether the tempoary html file has the right title
    with open(tmp_html_file) as html_file:
        tmp_html = BeautifulSoup(html_file.read(), features="html.parser")
    no_tag_err_msg = f"No h1 tag after title update on {note_file}"
    assert tmp_html.find("h1"), no_tag_err_msg
    mismatch_err_msg = f"H1 tag's string incorrect after title update on {note_file}"
    assert tmp_html.find("h1").text.strip() == updated_title, mismatch_err_msg
    # check whether the temporary html has the right body
    tmp_html.find("h1").decompose()
    assert tmp_html.prettify() == original_body
    # cleanup
    shutil.rmtree(notes_dir)


def _note_update_title_untitled(note_file: Path):
    # setup
    notes_dir = Path(tempfile.mkdtemp())
    with open(note_file) as file:
        note_path = create_note_file(file.read(), notes_dir)
    note = Note(note_path)
    original_title = note.get_title()
    original_body = BeautifulSoup(note.get_body(), features="html.parser").prettify()
    note.update_title("")
    # check whether the temporary html file exists
    tmp_html_file = Path(notes_dir, ".tmp", f"{note.last_update}.html")
    assert tmp_html_file.exists(), "Temporary html file doesn't exist after update"
    # check whether the original disk note file unmodified
    original_note = Note(note_path, cache=False)
    modified_err_msg = f"Original disk file got modifed while updating {note_file}"
    assert original_note.get_title() == original_title, modified_err_msg
    original_note_body = BeautifulSoup(original_note.get_body(), features="html.parser").prettify()
    assert original_note_body == original_body, modified_err_msg
    # check whether newly loaded note has the updated title
    new_note = Note(note_path)
    assert new_note.get_title() == "", modified_err_msg
    assert new_note.get_body() == original_body, modified_err_msg
    # check whether the tempoary html file has the right title (no h1 tag)
    with open(tmp_html_file) as html_file:
        tmp_html = BeautifulSoup(html_file.read(), features="html.parser")
    tag_err_msg = f"H1 tag still exist after clearing title in {note_file}"
    assert not tmp_html.find("h1"), tag_err_msg
    # check whether the temporary html has the right body
    assert tmp_html.prettify() == BeautifulSoup(original_body, features="html.parser").prettify()
    # cleanup
    shutil.rmtree(notes_dir)


def test_note_update_title():
    """
    There are several tricky cases for updating a note's title:
    1. From untitled to titled
    2. From titled to titled
    3. From titled to untitled
    """
    # Update with a title
    for note_file in NOTE_HTML_FILES:
        _note_update_title_titled(note_file)
    # Remove the title from the note if exist
    for note_file in NOTE_HTML_FILES:
        _note_update_title_untitled(note_file)


def test_note_save():
    """
    While a note is being updated, a temporary html file with the latest content is generated.
    During the editing, the original note file keeps unmodified.
    After the note is saved by the user, the temporary and original file should be delted.
    A new note file with the update stamp and updated content should appear in the notes folder.
    """
    # setup
    note_file = "./notes/title_heading_content.html"
    notes_dir = Path(tempfile.mkdtemp())
    with open(note_file) as file:
        note_path = create_note_file(file.read(), notes_dir)
    note = Note(note_path)
    # update the note with a different title
    updated_title = "updated title by test_note_save"
    note.update_title(updated_title)
    # check whether the temporary html file exists
    tmp_html_file = Path(notes_dir, ".tmp", f"{note.last_update}.html")
    assert tmp_html_file.exists(), "Temporary html file doesn't exist after update"
    time.sleep(1.5)  # sleep 1.5 second, otherwise the saved note will have the same timestamp
    note.save()
    # check whether the temporary html file is deleted
    assert not tmp_html_file.exists(), "Temporary html file still exist after save"
    # check wehther the old note file is deleted
    assert not note_path.exists(), "Old note still exist after save"
    # check whether the new note file is generated
    files = glob.glob(f"{os.fspath(notes_dir)}/*.json")
    assert len(files) == 1, "New note json file isn't created"
    # check whether the new note file has the right
    updated_note = Note(files[0])
    assert updated_note.get_title() == updated_title, "Saved note doesn't have the right title"
    # cleanup
    shutil.rmtree(notes_dir)


def test_note_reset():
    """
    When a note got updated, a temporary html file with the latest content is generated.
    If the quit NBX without saving the file, the temporary file remains there.
    When the note is reset explicitly, the temporary file should be delted.
    """
    # setup
    note_file = "./notes/title_heading_content.html"
    notes_dir = Path(tempfile.mkdtemp())
    with open(note_file) as file:
        note_path = create_note_file(file.read(), notes_dir)
    note = Note(note_path)
    # update the note with a different title
    updated_title = "updated title by test_note_save"
    note.update_title(updated_title)
    # check whether the temporary html file exists
    tmp_html_file = Path(notes_dir, ".tmp", f"{note.last_update}.html")
    assert tmp_html_file.exists(), "Temporary html file doesn't exist after update"
    note.reset()
    # check whether the temporary html file is deleted
    assert not tmp_html_file.exists(), "Temporary html file still exist after save"
    # check whether the note has the original content
    new_note = Note(note_path)
    err_msg = "Note doesn't have the same title after reset"
    assert new_note.get_title() == "Title: title_heading_content", err_msg
    # cleanup
    shutil.rmtree(notes_dir)


def test_note_archive():
    """
    When a note is archived, the note's json file shouldn't appear in the notes directory anymore.
    Instead, it should be moved to the archive folder.
    """
    # setup
    note_file = "./notes/title_heading_content.html"
    notes_dir = Path(tempfile.mkdtemp())
    with open(note_file) as file:
        note_path = create_note_file(file.read(), notes_dir)
    note = Note(note_path)
    note.archive()
    # check whether the note json file is deleted
    assert not note_path.exists(), "Note json file still exist after archive"
    # check wehther the note json file is moved to the archive folder
    archive_note_path = Path(notes_dir, "archive", f"{note.last_update}.json")
    assert archive_note_path.exists(), "Archived json file doesn't exist after archive"
    # check whether the archived note still could be loaded correctly
    archived_note = Note(archive_note_path)
    err_msg = "Archived note could not be loaded correctly"
    assert archived_note.get_title() == "Title: title_heading_content", err_msg
    # cleanup
    shutil.rmtree(notes_dir)


def test_note_construction():
    for idx, note_file in enumerate(NOTE_HTML_FILES):
        # setup
        notes_dir = Path(tempfile.mkdtemp())
        with open(note_file) as file:
            note_path = create_note_file(file.read(), notes_dir)
        note = Note(note_path)
        # check whether title is correctly constructed
        title_err_msg = f"Title is not correctly constructed for note {note_file}"
        assert note.get_title() == NOTE_TITLES[idx], title_err_msg
        # check whether headline is correctly constructed
        headline_err_msg = f"Headline is not correctly constructed for note {note_file}"
        assert note.get_headline() == NOTE_HEADLINES[idx], headline_err_msg
        # check whether the sections are correctly constructed
        sections_err_msg = f"Sections are not correctly constructed for note {note_file}"
        assert len(note.get_sections()) == NOTE_SECTION_CNTS[idx], sections_err_msg
        # cleanup
        shutil.rmtree(notes_dir)


if __name__ == "__main__":

    # given a correct html input, nbx.note's api should:
    # extract the correct corresponding content
    test_note_construction()

    # given a specific operation on a note, nbx.note's api should:
    # apply the desired changes and reflect the changes to disk
    test_note_update_body()
    test_note_update_title()
    test_note_save()
    test_note_reset()
    test_note_archive()



import os

from pathlib import Path
from typing import List, Tuple

from nbx import config
from nbx.note import Note, Record

SEARCH_OPTIONS = [
    "regex",
    "whoosh",
    "faiss",
    "pinecone"
]


def search_notes(
    notes: List[Note],
    query: str,
    method: str = "regex"
) -> Tuple[List[Record], List[Note]]:
    # Query are searched in lowercased
    query = query.lower().strip()

    def fallback_func():
        print("[Warning] Falling back to regex search")
        return search_notes(notes, query, "regex")

    # MOCK & Regex are simple
    if os.getenv("DEBUG", "False") == "True":
        from .mock import find_match
        return find_match(notes, query)
    elif method == "regex":
        from .regex import find_match
        return find_match(notes, query)

    # Search with index
    nbx_config = config.load_config()
    notes_dir = nbx_config["notes_dir"]
    if method == "whoosh":
        from .whoosh import find_match
        index_dir = Path(notes_dir, "whoosh")
        return find_match(query, index_dir, notes_dir)
    elif method == "faiss":
        from .faiss import find_match
        index_dir = Path(notes_dir, "faiss")
        return find_match(query, index_dir, notes_dir)
    elif method == "txtai":
        from .txtai import find_match
        find_match(notes, query)
    elif method == "pinecone":
        from .pinecone import find_match
        pinecone_token = nbx_config.get("pinecone_token", "")
        return find_match(query, pinecone_token, notes_dir, fallback_func)
    elif method.startswith("openai"):
        from .openai import find_match
        model = method.split("-", 1)[1]

        def openai_search_func(*args, **kwargs):
            find_match(*args, *kwargs, model=model)

        return openai_search_func(notes, query)
    else:
        return [], []


def index_notes(notes: List[Note], method: str = "regex") -> None:
    print(f"[Info] Indexing using {method}")
    if method == "pinecone":
        from .pinecone import index_notes
        pinecone_token = config.load_pinecone_token()
        index_notes(notes, pinecone_token)
    elif method == "whoosh":
        from .whoosh import index_notes
        notes_dir = config.load_config()["notes_dir"]
        index_notes(notes, Path(notes_dir, "whoosh"))
    elif method == "faiss":
        from .faiss import index_notes
        notes_dir = config.load_config()["notes_dir"]
        index_notes(notes, Path(notes_dir, "faiss"))
    return



import logging
from pathlib import Path

from .cli import cli

log_dir = Path(Path.home(), ".nbx")
log_dir.mkdir(parents=True, exist_ok=True)
log_file = Path(log_dir, "nbx.log")
log_file.touch(exist_ok=True)

logging.basicConfig(
    filename=log_file,
    level=logging.WARNING,
    format='[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
    datefmt='%H:%M:%S'
)

__all__ = [cli]



import base64
import glob
import json
import os
import pathlib
import psutil
import re
from typing import Any, Dict
from pathlib import Path


def load_json(file_path: pathlib.Path) -> Dict:
    with open(file_path, "r") as file:
        try:
            json_obj = json.load(file)
            return json_obj
        except json.JSONDecodeError as e:
            if os.stat(file_path).st_size == 0:
                # Empty file
                return {}
            else:
                raise e


def save_json_obj(file_path: pathlib.Path, json_obj: Dict) -> None:
    with open(file_path, "w+") as json_file:
        json.dump(json_obj, json_file, sort_keys=True)
        json_file.flush()


def set_json_value(file_path: pathlib.Path, key: str, value: Any) -> None:
    # key needs to follow python naming convention, such as trial_id
    config = {}
    if file_path.exists():
        with open(file_path, "r") as json_file:
            try:
                config = json.load(json_file)
            except json.JSONDecodeError as e:
                if os.stat(file_path).st_size == 0:
                    # Empty file
                    config = {}
                else:
                    raise e
    config[key] = value
    with open(file_path, "w+") as json_file:
        json.dump(config, json_file)
        json_file.flush()


def load_json_value(file_path: pathlib.Path, key: str, default_value: Any = "") -> Any:
    with open(file_path, "r") as json_file:
        try:
            config = json.load(json_file)
        except json.JSONDecodeError as e:
            if os.stat(file_path).st_size == 0:
                # Empty file
                return default_value
            else:
                raise e
        if key not in config:
            return None
        return config[key]


def terminate_process():
    current_process = psutil.Process(os.getpid())
    current_process.terminate()


def verify_openai_token(token: str) -> str:
    import openai
    openai.api_key = token
    try:
        openai.Completion.create(
            model="text-ada-001",
            prompt="Hello",
            temperature=0,
            max_tokens=10,
            top_p=1,
            frequency_penalty=0.5,
            presence_penalty=0
        )
        return "OK"
    except Exception as e:
        return str(e)


def verify_pinecone_token(token: str) -> str:
    return "OK"


def save_base64(img_data, file_path: pathlib.Path):
    with open(file_path, "wb") as fh:
        fh.write(base64.decodebytes(bytes(img_data, encoding="utf-8")))


def lang_detect(text: str):
    # # korean
    # if re.search("[\uac00-\ud7a3]", texts):
    #     return "ko"
    # # japanese
    # if re.search("[\u3040-\u30ff]", texts):
    #     return "ja"
    # chinese
    if re.search("[\u4e00-\u9FFF]", text):
        return "zh"
    return None


def count_zh_char(text: str) -> int:
    res = len(re.findall("[\u4e00-\u9FFF]", text))
    res += len(re.findall("，", text))
    res += len(re.findall("。", text))
    res += len(re.findall("：", text))
    res += len(re.findall("？", text))
    return res


def clean_folder(folder: Path):
    files = glob.glob(f"{os.fspath(folder)}/*")
    for f in files:
        os.remove(f)



def test_app_save_note():
    pass


def test_app_update_note_title():
    pass


def test_app_update_note_body():
    pass


if __name__ == "__main__":
    test_app_save_note()
    test_app_update_note_title()
    test_app_update_note_body()



import os
import shutil
from abc import ABC
from datetime import datetime
from pathlib import Path
from typing import List

from bs4 import BeautifulSoup

from nbx.util import save_json_obj, load_json


class Record(ABC):
    def __init__(
        self,
        id: str,
        heading: str,
        headingType: str,
        headingIdx: int,
        content: str,
        filename: str,
    ) -> None:
        super().__init__()
        self.id = id
        self.heading = heading
        # heading's corresponding html tag type
        self.headingType = headingType
        # heading index based on its headingType
        self.headingIdx = headingIdx
        self.content = content
        self.filename = filename


class Note(ABC):

    def __init__(self, file: Path, cache: bool = True) -> None:
        super().__init__()
        self.file = file
        self.logs = {}
        self.last_update = ""
        self.html = ""
        self.notes_dir = os.path.dirname(file)
        self.has_changed = False
        self._load_file(file, cache)

    def _load_file(self, note_file: Path, cache: bool) -> None:
        # load the main note
        note = load_json(note_file)
        self.logs = note["content"]
        self.last_update = note["lastUpdate"]
        self.html = self.logs[self.last_update]
        # load the temporary note if exist
        tmp_note_file = Path(os.path.dirname(note_file), ".tmp", f"{self.last_update}.html")
        if tmp_note_file.exists() and cache:
            self.has_changed = True
            with open(tmp_note_file) as tmp_note_file:
                self.html = tmp_note_file.read()

    def _save_tmp_note(self):
        tmp_notes_dir = Path(self.notes_dir, ".tmp")
        if not tmp_notes_dir.exists():
            tmp_notes_dir.mkdir(parents=True)
        tmp_note_file_path = Path(tmp_notes_dir, f"{self.last_update}.html")
        with open(tmp_note_file_path, "w+") as tmp_note_file:
            tmp_note_file.write(self.html)

    def save(self) -> None:
        if not self.has_changed:
            return
        timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        last_update = self.last_update
        self.last_update = timestamp
        self.logs[timestamp] = self.html
        # save the note to disk
        note_json = {"lastUpdate": timestamp, "content": self.logs}
        note_file = Path(self.notes_dir, f"{timestamp}.json")
        save_json_obj(note_file, note_json)
        # remove tmporary note
        tmp_note_file = Path(self.notes_dir, ".tmp", f"{last_update}.html")
        if tmp_note_file.exists():
            os.remove(tmp_note_file)
        # remove old note
        old_note = Path(self.notes_dir, f"{last_update}.json")
        if old_note.exists():
            os.remove(old_note)
        self.has_changed = True

    def archive(self) -> None:
        archive_notes_dir = Path(self.notes_dir, "archive")
        archive_notes_dir.mkdir(parents=True, exist_ok=True)
        src_path = Path(self.notes_dir, f"{self.last_update}.json")
        dst_path = Path(archive_notes_dir, f"{self.last_update}.json")
        shutil.move(src_path, dst_path)

    def reset(self) -> None:
        self.html = self.logs[self.last_update]
        # remove temporary note
        tmp_note_file = Path(self.notes_dir, ".tmp", f"{self.last_update}.html")
        if tmp_note_file.exists():
            os.remove(tmp_note_file)

    def update_title(self, title: str) -> None:
        self.has_changed = True
        doc = BeautifulSoup(self.html, features="html.parser")
        components = doc.find_all(recursive=False)
        # remove the old title if exist
        if len(components) > 0 and components[0].name == "h1":
            components[0].decompose()
        # add the new title
        if title:
            if "<h1>" in title:
                title_component = BeautifulSoup(title, features="html.parser")
            else:
                title_component = BeautifulSoup(f"<h1>{title}</h1>", features="html.parser")
            # add title if it is not empty
            if title_component.text.strip():
                doc.insert(0, title_component)
        self.html = doc.prettify()
        self._save_tmp_note()

    def update_body(self, body_html: str) -> None:
        self.has_changed = True
        doc = BeautifulSoup(self.html, features="html.parser")
        components = doc.find_all(recursive=False)
        body_component = BeautifulSoup(body_html, features="html.parser")
        # if the doc contains a title
        if len(components) > 0 and components[0].name == "h1":
            title_component = components[0]
            body_component.insert(0, title_component)
        self.html = body_component.prettify()
        self._save_tmp_note()

    def get_headline(self) -> str:
        note_content = BeautifulSoup(self.html, features="html.parser")
        components = note_content.find_all(recursive=False)
        if len(components) == 0:
            return ""
        return components[0].text.strip()

    def get_title(self) -> str:
        note_content = BeautifulSoup(self.html, features="html.parser")
        if note_content.find("h1"):
            return note_content.find("h1").text.strip()
        return ""

    def get_body(self) -> str:
        note_content = BeautifulSoup(self.html, features="html.parser")
        components = note_content.find_all(recursive=False)
        if len(components) == 0:
            return ""
        if components[0].name == "h1":  # A note starts with a H1 heading
            components[0].decompose()
            return note_content.prettify()
        return self.html

    def get_tags(self) -> List[str]:
        return [self.last_update[:10]]  # Date like [2022-02-22] as the default tag

    def get_sections(self) -> List[Record]:
        heading_index_map = {"h1": 0, "h2": 0, "p": 0}
        note_content = BeautifulSoup(self.html, features="html.parser")
        components = note_content.find_all(recursive=False)
        if len(components) == 0:
            return []
        # Extract records from the note
        records = []
        if components[0].name == "h1":  # A note starts with a H1 heading
            heading = components[0].text.strip()
            components = components[1:]
            content = heading  # If the h1 heading has no content, then use the heading
            if len(components) > 0 and components[0].name == "p":
                content = components[0].text.strip()
                components = components[1:]
                heading_index_map["p"] = heading_index_map.get("p") + 1
            record = Record(
                f"{self.last_update}#{len(records)}",
                heading=heading,
                headingType="h1",
                headingIdx=heading_index_map.get("h1"),
                content=content,
                filename=os.path.basename(self.file),
            )
            heading_index_map["h1"] = heading_index_map.get("h1") + 1
            records.append(record)
        elif components[0].name != "h2":  # A note starts with a content, i.e. not section heading
            content = components[0].text.strip()
            components = components[1:]
            record = Record(
                f"{self.last_update}#{len(records)}",
                heading=content,
                headingType="p",
                headingIdx=heading_index_map.get("p"),
                content=content,
                filename=os.path.basename(self.file),
            )
            heading_index_map["p"] = heading_index_map.get("p") + 1
            records.append(record)
        # Main body of the note: <h2>heading</h2><p>content</p>
        idx = 0
        record_heading = ""
        record_content = ""
        while len(components) > 0:
            component = components[0]
            components = components[1:]
            if component.name == "h2":
                if idx != 0:  # save previous reocrd
                    record = Record(
                        f"{self.last_update}#{len(records)}",
                        heading=record_heading.strip(),
                        headingType="h2",
                        headingIdx=heading_index_map.get("h2"),
                        content=record_content.strip(),
                        filename=os.path.basename(self.file),
                    )
                    records.append(record)
                    heading_index_map["h2"] = heading_index_map.get("h2") + 1
                record_heading = component.text.strip()
                record_content = ""
                idx += 1
            else:
                content = component.text.strip()
                record_content += content + "\n"
        if record_heading:
            record = Record(
                f"{self.last_update}#{len(records)}",
                heading=record_heading,
                headingType="h2",
                headingIdx=heading_index_map.get("h2"),
                content=record_content,
                filename=os.path.basename(self.file),
            )
            records.append(record)
            heading_index_map["h2"] = heading_index_map.get("h2") + 1
        return records


def create_note_file(note_html: str, notes_dir: Path, note_filename: str = "") -> Path:
    notes_dir = Path(notes_dir)
    if not notes_dir.exists():
        notes_dir.mkdir(parents=True)
    timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    if not note_filename:
        note_filename = timestamp
    note_file_path = Path(notes_dir, f"{note_filename}.json")
    save_json_obj(note_file_path, {
        "lastUpdate": timestamp,
        "content": {
            timestamp: note_html
        }
    })
    return note_file_path



import uuid
from typing import List, Tuple

from faker import Faker

from nbx.note import Note, Record


def find_match(*args, **kwargs) -> Tuple[List[Record], List[Note]]:
    Faker.seed(0)
    fake = Faker()

    records = []
    notes = []
    for _ in range(kwargs.get("num", 20)):
        records.append(Record(str(uuid.uuid4()), fake.text()[:8], fake.text()))
        notes.append(fake.file_path())

    return records, notes



import re
from typing import List, Tuple

from nbx.note import Note, Record


def appear_in_record(query: str, record: Record) -> bool:
    content = record.content
    heading = record.heading
    return re.search(query, content) or re.search(query, heading)


def find_match(notes: List[Note], query: str) -> Tuple[List[Record], List[Note]]:
    res_records = []
    res_notes = []
    for note in notes:
        for record in note.get_sections():
            if appear_in_record(query, record):
                res_records.append(record)
                res_notes.append(note)
    return res_records, res_notes



import logging
import os
import pathlib
import webbrowser
from pathlib import Path
from threading import Timer
from typing import Any, Dict, List

import uvicorn
from fastapi import FastAPI, WebSocket, WebSocketDisconnect, HTTPException
from fastapi.responses import RedirectResponse, FileResponse
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware

from nbx import nbx, util, config
from nbx.note import Note

FRONTEND_SESSION_STATE = {
    # See all page options in ./frontend/src/App.svelte
    "routing": {"page": "", "params": {}},
    # EditNote related
    "relevantRecords": [],
    "uploadedImageUrl": "",
    "noteTitle": "",
    "noteContent": "",
    # EditInsight related
    "insightGraph": {"nodes": [], "edges": [], "selectedNodes": []},
    "statementHtmls": [],
    # ListNotes related
    "selectedNoteIdx": -1,
    # Connection related
    "connectedToBackend": True,
}

BACKEND_SESSION_STATE = {
    # EditNote related
    "note": None,
    # ListNotes related
}

logger = logging.getLogger(__name__)

####################################################################################################
# FastAPI Configurations

HOST = "localhost"
PORT = 8000

app = FastAPI()

# Allow CORS from svelte frontend to fastapi backend
origins = [
    f"http://{HOST}",
    f"http://{HOST}:{PORT}",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

pwd = pathlib.Path(__file__).parent.resolve()
app.mount(
    "/app",
    StaticFiles(directory=Path(pwd, "frontend/dist/"), html=True),
    name="static",
)

####################################################################################################
# Logics for handling data sent by the UI frontend


def handle_save_note():
    note = BACKEND_SESSION_STATE["note"]
    has_changed = True
    if note:
        has_changed = note.has_changed
        note.save()
    else:
        full_content = FRONTEND_SESSION_STATE["noteContent"]
        if FRONTEND_SESSION_STATE["noteTitle"] != "<h1></h1>":
            full_content = FRONTEND_SESSION_STATE["noteTitle"] + "\n" + full_content
        nbx.create_note(full_content)
    if has_changed:
        nbx.index_all_notes(background=True)
    util.terminate_process()


def handle_exit():
    note = BACKEND_SESSION_STATE["note"]
    if note:
        note.reset()
    util.terminate_process()


def handle_update_note(data):
    global FRONTEND_SESSION_STATE
    if "data" not in data:
        logger.warning("Invalid update_note websocket: No data specified", data)
        return

    input = data["data"]["input"]
    content = data["data"]["content"]
    update_type = data["data"]["updateType"]
    if update_type == "noteContent":
        FRONTEND_SESSION_STATE["noteContent"] = content
    if update_type == "noteTitle":
        FRONTEND_SESSION_STATE["noteTitle"] = content

    note = BACKEND_SESSION_STATE["note"]
    if note:
        if update_type == "noteContent":
            note.update_body(content)
        if update_type == "noteTitle":
            note.update_title(content)
    # find relevant note abstracts as prompts for the user to select
    relevantRecords = nbx.find_relevant_records(input)
    FRONTEND_SESSION_STATE["relevantRecords"] = relevantRecords


def handle_show_note(data):
    global FRONTEND_SESSION_STATE
    if "data" not in data:
        logger.warning("Invalid update_note websocket: No data specified", data)
        return

    filename = data["data"]["filename"]
    nbx_config = config.load_config()
    note = Note(Path(nbx_config.get("notes_dir"), filename))
    note_title = "<h1>" + note.get_title() + "</h1>"  # get_title returns pure text without h1 tag
    note_content = note.get_body()
    FRONTEND_SESSION_STATE["routing"]["params"] = {
        "noteTitle": note_title,
        "noteContent": note_content,
        "logs": note.logs,
        "filename": filename,
    }
    FRONTEND_SESSION_STATE["noteTitle"] = note_title
    FRONTEND_SESSION_STATE["noteContent"] = note_content
    BACKEND_SESSION_STATE["note"] = note


def handle_upload_image(data):
    global FRONTEND_SESSION_STATE

    if "data" not in data:
        logger.warning("Invalid upload_image websocket: No data specified", data)
        return
    image = data["data"]["image"]
    image_name = nbx.add_image(image)
    FRONTEND_SESSION_STATE[
        "uploadedImageUrl"
    ] = f"http://{HOST}:{PORT}/images/{image_name}"


def handle_data(data):
    if "type" not in data:
        logger.warning("Invalid websocket data: No type specified", data)
        return
    data_type = data["type"]

    if data_type == "exit":
        handle_exit()
    elif data_type == "save_note":
        handle_save_note()
    elif data_type == "update_note":
        handle_update_note(data)
    elif data_type == "upload_image":
        handle_upload_image(data)
    elif data_type == "show_note":
        handle_show_note(data)


####################################################################################################
# FastAPI Endpoints


@app.get("/images/{filename}")
async def images(filename: str) -> FileResponse:
    image_path = nbx.get_image_path(filename)
    if not image_path.exists():
        raise HTTPException(status_code=404, detail="Item not found")
    return FileResponse(image_path)


@app.get("/data")
async def data():
    return FRONTEND_SESSION_STATE


@app.websocket("/stream")
async def stream(websocket: WebSocket):
    await websocket.accept()
    try:
        while True:
            data = await websocket.receive_json()
            handle_data(data)
            await websocket.send_json(FRONTEND_SESSION_STATE)
    except WebSocketDisconnect:
        logger.info("connection closed")


@app.get("/")
async def root():
    return RedirectResponse(url="app")


def add_note(host: str = HOST, port: int = PORT) -> None:
    global FRONTEND_SESSION_STATE
    FRONTEND_SESSION_STATE["routing"] = {"page": "EditNote", "params": {}}
    webbrowser.open_new(f"http://{host}:{port}")
    uvicorn.run(app, host=host, port=port)


def edit_note(
    note: Note, host: str = HOST, port: int = PORT, read_only: bool = False
) -> None:
    global FRONTEND_SESSION_STATE
    global BACKEND_SESSION_STATE
    note_title = "<h1>" + note.get_title() + "</h1>"  # get_title returns pure text without h1 tag
    note_content = note.get_body()
    # @TODO This param is a redundant info as it is only needed to initialize UI components
    # And also the info is already recorded in the FRONTEND_SESSION_STATE
    add_note_params = {
        "noteTitle": note_title,
        "noteContent": note_content,
        "logs": note.logs,
        "filename": os.path.basename(note.file),
    }
    FRONTEND_SESSION_STATE["routing"] = {"page": "EditNote", "params": add_note_params}
    FRONTEND_SESSION_STATE["connectedToBackend"] = not read_only
    FRONTEND_SESSION_STATE["noteTitle"] = note_title
    FRONTEND_SESSION_STATE["noteContent"] = note_content
    BACKEND_SESSION_STATE["note"] = note
    webbrowser.open_new(f"http://{host}:{port}")
    if read_only:
        Timer(1, util.terminate_process).start()  # kill the python backend in 1 second
    uvicorn.run(app, host=host, port=port)


def list_notes(
    note_infos: List[Dict[str, Any]], host: str = HOST, port: int = PORT
) -> None:
    global FRONTEND_SESSION_STATE
    FRONTEND_SESSION_STATE["routing"] = {
        "page": "ListNotes",
        "params": {"noteInfos": note_infos},
    }
    webbrowser.open_new(f"http://{host}:{port}")
    uvicorn.run(app, host=host, port=port)



from bs4 import NavigableString
import openai
from typing import List
from nbx import nbx


def construct_prompt(query, document):
    return "<|endoftext|>{document}\n\n---\n\nThe above passage is related to: {query}".format(
        document=document, query=query
    )


def get_score(context, query, log_probs, text_offsets) -> float:
    SCORE_MULTIPLIER = 100.0

    log_prob = 0
    count = 0
    cutoff = len(context) - len(query)

    for i in range(len(text_offsets) - 1, 0, -1):
        log_prob += log_probs[i]
        count += 1

        if text_offsets[i] <= cutoff and text_offsets[i] != text_offsets[i - 1]:
            break

    return log_prob / float(count) * SCORE_MULTIPLIER


def openai_search(query, documents, engine):
    prompts = [construct_prompt(query, doc) for doc in [""] + documents]

    resps = openai.Completion.create(
        model=engine,
        prompt=prompts,
        temperature=1.0,
        top_p=1.0,
        max_tokens=0,
        logprobs=0,
        n=1,
        echo=True,
    )

    resps_by_index = {choice["index"]: choice for choice in resps["choices"]}

    scores = [
        get_score(
            prompts[i],
            query,
            resps_by_index[i]["logprobs"]["token_logprobs"],
            resps_by_index[i]["logprobs"]["text_offset"],
        )
        for i in range(len(prompts))
    ]

    # Process results
    scores = [score - scores[0] for score in scores][1:]

    return [
        {
            "object": "search_result",
            "document": document_idx,
            "score": round(score, 3),
        }
        for document_idx, score in enumerate(scores)
    ]


def find_match(text_objs: List[NavigableString], query: str, model: str) -> List[NavigableString]:
    token = nbx.load_openai_token()
    if token != "":
        openai.api_key = token
    else:
        print("[Error] No openai token")
        return []

    # search
    MATCH_THRESHOLD_GPT3 = 200
    res = []
    search_results = openai_search(query, text_objs, model)
    for search_result in search_results:
        if search_result['score'] >= MATCH_THRESHOLD_GPT3:
            res.append(text_objs[search_result['document']])
    return res



import os
from datetime import datetime
from pathlib import Path
from typing import List, Tuple

import faiss
import pandas as pd
import numpy as np
from sentence_transformers import SentenceTransformer

from nbx import util
from nbx.note import Note, Record

MODEL_CKPT = "flax-sentence-embeddings/all_datasets_v3_mpnet-base"


def find_match(
    query: str,
    index_dir: Path,
    notes_dir: Path,
    top_k: int = 10
) -> Tuple[List[Record], List[Note]]:
    if not index_dir.exists():
        print("No index found, please index the notebase first: nbx index")
        return [], []

    # get the index with the latest timestamp
    index_files = index_dir.glob("*.index")
    index_files = sorted(index_files, reverse=True)
    if len(index_files) == 0:
        print("No index found, please index the notebase first: nbx index")
        return [], []

    # check index_file and the corresponding info json
    index_file = index_files[0]
    index_timestamp = index_file.name.split(".")[0]
    index_info_file = Path(index_dir, f"{index_timestamp}.json")
    if not index_info_file.exists():
        print("Corrupted index, please re-index the notebase first: nbx index")
        return [], []

    # load faiss index
    index_path = os.fspath(index_file)
    index = faiss.read_index(index_path)

    # create the search vector for the query
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    encoder = SentenceTransformer(MODEL_CKPT)
    vector = encoder.encode(query)
    _vector = np.array([vector])
    faiss.normalize_L2(_vector)
    top_k = min(top_k, index.ntotal)
    distances, ann = index.search(_vector, k=top_k)

    # get record(section) ids
    index_info = util.load_json(index_info_file)
    res_records = []
    res_notes = []
    loaded_notes = {}
    loaded_records = {}
    for ann_idx in ann[0]:
        record_id = index_info[str(ann_idx)]
        note_id = record_id.split("#")[0]
        record_idx = int(record_id.split("#")[1])
        if note_id not in loaded_notes:
            note = Note(Path(notes_dir, f"{note_id}.json"))
            loaded_notes[note_id] = note
            loaded_records[note_id] = note.get_sections()
        note = loaded_notes[note_id]
        record = loaded_records[note_id][record_idx]
        res_records.append(record)
        res_notes.append(note)

    return res_records, res_notes


def index_notes(notes: List[Note], index_dir: Path = "") -> None:
    # create index_dir if doesn't exist
    index_dir.mkdir(parents=True, exist_ok=True)

    # convert notes content into a dataframe
    data = []
    for note in notes:
        for record in note.get_sections():
            data.append([record.content, record.id])
    df = pd.DataFrame(data, columns=["text", "id"])

    # no need to index if no note exists yet
    if len(data) == 0:
        return

    # create vector from the text
    encoder = SentenceTransformer(MODEL_CKPT)
    vectors = encoder.encode(df["text"])

    # build the faiss index from the vectors
    vector_dimension = vectors.shape[1]
    index = faiss.IndexFlatL2(vector_dimension)
    faiss.normalize_L2(vectors)
    index.add(vectors)

    # save the index to disk
    timestamp = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
    index_path = os.fspath(Path(index_dir, f"{timestamp}.index"))
    faiss.write_index(index, index_path)
    # save some additional info
    info_path = os.fspath(Path(index_dir, f"{timestamp}.json"))
    util.save_json_obj(info_path, df["id"].to_dict())
